@@ -14,7 +14,7 @@ from .common import *
 import requests
 from itertools import cycle
 
-__version__ = '2.0.14'
+__version__ = '2.1.0'
 
 if NUMPY_INSTALLED:
     from .points_serializers import np_array
@@ -419,7 +419,7 @@ class Connection(object):
             return None
         return data.get('fields', {})
 
-    def get_reported_dates(self, series_query=None, job_id=None, processor=None, df=None, dt=None):
+    def get_reported_dates(self, series_query=None, job_id=None, processor=None, df=None, dt=None, mode='points'):
         """
         Retrieves reported dates for one of the following (use only one): series_query, job_id, processor.  Dates returned can be limited via df/dt.
 
@@ -428,10 +428,12 @@ class Connection(object):
         :param processor:  return reported dates for series written by that processor
         :param df: datetime to start the dates array
         :param dt: datetime to finish the dates array
+        :param mode: return either `points` or `fields` reported dates
         :return: list of reported dates
         """
         reported_dates = self.shooju_api.get_reported_dates(series_query=series_query, job_id=job_id,
-                                                            processor=processor, df=df, dt=dt).get('reported_dates') or []
+                                                            processor=processor, df=df, dt=dt,
+                                                            mode=mode).get('reported_dates') or []
         return list(map(milli_to_datetime, reported_dates))
 
     def register_job(self, description, notes="", batch_size=1,
@@ -631,9 +633,13 @@ class PandasFormatter(object):
             return [key_field]
 
 
-def _parse_dt(d, default):
+def _parse_dt(d, param_name, default):
     if d is not None and not isinstance(d, six.string_types):
-        d = to_milli(d)
+        try:
+            d = to_milli(d)
+        except ValueError:
+            raise ValueError("{} accepts only string, milliseconds since epoch, datetime.date,"
+                             " or datetime.datetime objects".format(param_name))
     elif d is None:
         d = default
     return d
@@ -682,7 +688,7 @@ class ShoojuApi(object):
                              retry_delay=kwargs.get('retry_delay'),
                              location=kwargs.get('location'),
                              hosts=kwargs.get('hosts'),
-                             proxies=kwargs.get('proxies'))
+                             requests_session=kwargs.get('requests_session'))
 
     def get_single_series(self, series_id=None, series_query=None, df=None, dt=None,
                           max_points=None, fields=None, snapshot_job_id=None, snapshot_date=None,
@@ -770,8 +776,8 @@ class ShoojuApi(object):
             deserializer_params['use_numpy'] = True
         url_params = {
             'date_format': 'milli',
-            'df': _parse_dt(df, 'MIN') if df is not None else None,
-            'dt': _parse_dt(dt, 'MAX') if dt is not None else None,
+            'df': _parse_dt(df, 'df', 'MIN') if df is not None else None,
+            'dt': _parse_dt(dt, 'dt', 'MAX') if dt is not None else None,
             'snapshot_job_id': snapshot_job_id,
             'snapshot_date': to_milli(snapshot_date) if snapshot_date else None,
             'reported_date': to_milli(reported_date) if reported_date else None,
@@ -960,8 +966,8 @@ class ShoojuApi(object):
         params = {
             'query': query,
             'date_format': 'milli',
-            'df': _parse_dt(date_start, 'MIN'),
-            'dt': _parse_dt(date_finish, 'MAX'),
+            'df': _parse_dt(date_start, 'date_start', 'MIN'),
+            'dt': _parse_dt(date_finish, 'date_finish', 'MAX'),
             'max_points': max_points,
         }
 
@@ -997,7 +1003,8 @@ class ShoojuApi(object):
         return self.client.get(self.API_SERIES, deserializer=options.series_serializer, params=params,
                                deserializer_params=deserializer_params)
 
-    def get_reported_dates(self, series_query=None, job_id=None, processor=None, df=None, dt=None):
+    def get_reported_dates(self, series_query=None, job_id=None, processor=None,
+                           df=None, dt=None, mode='points'):
         """
         Retrieves reported dates.
 
@@ -1006,14 +1013,17 @@ class ShoojuApi(object):
         :param processor:  return reported dates for series written by that processor
         :param df: datetime to start the dates array
         :param dt: datetime to finish the dates array
+        :param mode: return either `points` or `fields` reported dates
         :return: list of reported dates
         """
-        assert series_query or job_id or processor, 'At least one of series_query, job_id or processor parameters required'
+        assert series_query or job_id or processor, \
+            'At least one of series_query, job_id or processor parameters required'
         params = {
             'series_query': series_query,
             'job_id': job_id,
             'processor': processor,
-            'date_format': 'milli'
+            'date_format': 'milli',
+            'mode': mode
         }
         if df is not None:
             params['df'] = to_milli(df)
@@ -1059,7 +1069,7 @@ class Client(object):
     ALLOWED_METHODS = ('get', 'post', 'delete')  # list of allowed HTTP methods
 
     def __init__(self, server, user=None, password=None, base_path='/api/1',
-                 retries=None, retry_delay=None, location=None, hosts=None, proxies=None):
+                 retries=None, retry_delay=None, location=None, hosts=None, requests_session=None):
         """
         REST Client
 
@@ -1070,7 +1080,7 @@ class Client(object):
         :param retry_delay: time between api call attempts (seconds)
         :param location: Client location name
         :param hosts: Array of Shooju API endpoints. Used when access by ip, or different host name.
-        :param proxies: `requests` compatible proxy information to pass API requests through
+        :param requests_session: session object of requests.Session type to use as transport
         """
         if not hosts:
             hosts = [server]
@@ -1082,7 +1092,7 @@ class Client(object):
         self._retry_delay = retry_delay or 3.
         self._location = location
         self._methods = dict()
-        self._proxies = proxies
+        self._session = requests_session
         for m in self.ALLOWED_METHODS:
             self._methods[m] = functools.partial(self._call, m)
             # copy these attributes to make it look like original function (fix functool.wraps  + sj.raw.get issues)
@@ -1135,7 +1145,7 @@ class Client(object):
             payload = kwargs.get('data')
         headers['Sj-Receive-Format'] = deserializer.__name__
 
-        method = getattr(requests, method_name)
+        method = getattr(self._session if self._session is not None else requests, method_name)
 
         url = '{base}{path}'.format(base=next(self._endpoints), path=path)
         params = params or {}
@@ -1148,7 +1158,7 @@ class Client(object):
         transient_failure = False
         try:
             r = method(url, params=params or None, data=payload, headers=headers,
-                       files=files, auth=(self._user, self._password), timeout=timeout, proxies=self._proxies)
+                       files=files, auth=(self._user, self._password), timeout=timeout)
             if r.status_code != requests.codes.ok:
                 transient_failure = r.status_code in [502, 503, 504]
                 raise ConnectionError('Request failed. Error code %s' % r.status_code)
@@ -1257,11 +1267,13 @@ class BaseJob(object):
         # series_id will be the key and they will match to
         # {'fields':{},'points':{}} that !
         self._requests = OrderedDict()
+        self._queries = OrderedDict()
 
         self._pre_hooks = pre_hooks or []
         self._post_hooks = post_hooks or []
 
         self._remove = defaultdict(lambda: {'fields': False, 'points': False})
+        self._remove_query = defaultdict(lambda: {'fields': False, 'points': False})
         self._collision_retry_timeout = collision_retry_timeout
 
     @property
@@ -1277,6 +1289,19 @@ class BaseJob(object):
         """
         self.submit()
 
+    def _init_get_query_dict(self, query, delete=False, reported_date=None):
+        key = query, reported_date
+        if key not in self._queries:
+            self._queries.setdefault(key, {})
+            self._cur_batch_size += 1
+
+        if not delete and self._queries[key].get('op') != 'post':
+            self._queries[key] = {'fields': {}, 'points': {}, 'op': 'post'}
+        elif delete:
+            self._queries[key] = {'force': False, 'op': 'delete'}
+
+        return self._queries[key]
+
     def _init_get_series_dict(self, series, delete=False, reported_date=None):
         key = series, reported_date
         if key not in self._requests:
@@ -1297,9 +1322,64 @@ class BaseJob(object):
         if self._cur_batch_size >= self._batch_size:
             self.submit()
 
+    def write(self, series_query, fields=None, points=None, remove_others=None):
+        """
+        Writes the points and/or fields to the series_query. 
+        The series_query must represent 0 or 1 series (writing to more than one series is not supported).
+        To submit immediately, set batch_size to 1 or use submit().
+
+        :param series_query: series query
+        :param list points: list of shooju.Point
+        :param dict fields: fields dict
+        :param str remove_others: flag to remove all other points/points from the query.
+                                  Default is None, accepts: "points", "fields" and "all"
+        """
+        remove_opts = {
+            None: [False, False],
+            'points': [True, False],
+            'fields': [False, True],
+            'all': [True, True],
+        }
+        if remove_others not in remove_opts:
+            raise ValueError("remove_others should be one of None, all, points or fields")
+        remove_other_points, remove_other_fields = remove_opts[remove_others]
+        self._write(series_query, None, points, fields, remove_other_points, remove_other_fields)
+
+    def write_reported(self, series_query, reported_date, fields=None, points=None):
+        """
+        Writes points and/or fields to the series_query and associates them with reported_date.
+        The series_query must represent 0 or 1 series (writing to more than one series is not supported). 
+        To submit immediately, set batch_size to 1 or use submit().
+
+        :param series_query: series query
+        :param datetime.datetime reported_date: date reported
+        :param list points: list of shooju.Point
+        :param dict fields: fields dict
+        """
+        self._write(series_query, reported_date, points, fields, False, False)
+
+    def _write(self, series_query, reported_date, points, fields, remove_other_points, remove_other_fields):
+        request = self._init_get_query_dict(series_query, reported_date=reported_date)
+        if points:
+            for p in points:
+                request['points'].update(p.to_dict())
+
+        if fields:
+            request['fields'].update(fields)
+
+        if remove_other_points:
+            self._remove_query[series_query]['points'] = True
+
+        if remove_other_fields:
+            self._remove_query[series_query]['fields'] = True
+
+        self._submit_if_bulk()
+
     def put_points(self, series_id, points):
         """
-        Writes the points for the series_id.  To submit immediately, set batch_size to 1 or use submit().
+        DEPRECATED: Use write() method instead.
+
+        Writes the points for the series_id. To submit immediately, set batch_size to 1 or use submit().
 
         :param series_id: series id
         :param list points: list of shooju.Point
@@ -1312,7 +1392,9 @@ class BaseJob(object):
 
     def put_reported_points(self, series_id, reported_date, points):
         """
-        Writes the points for the series_id and associates them with reported_date.  To submit immediately, set batch_size to 1 or use submit().
+        DEPRECATED: Use write_reported() method instead.
+
+        Writes the points for the series_id and associates them with reported_date. To submit immediately, set batch_size to 1 or use submit().
 
         :param series_id: series id
         :param datetime.datetime reported_date: date reported
@@ -1326,7 +1408,9 @@ class BaseJob(object):
 
     def put_point(self, series_id, pt):
         """
-        Writes the point for the series_id.  To submit immediately, set batch_size to 1 or use submit().
+        DEPRECATED: Use write() method instead.
+
+        Writes the point for the series_id. To submit immediately, set batch_size to 1 or use submit().
 
         :param shooju.Point pt: point
         :param str series_id: series id
@@ -1338,12 +1422,16 @@ class BaseJob(object):
 
     def put_field(self, series_id, field_name, field_value):
         """
-        Writes the field_value under the field_name for the series_id.  To submit immediately, set batch_size to 1 or use submit().
+        DEPRECATED: Use write() method instead.
+
+        Writes the field_value under the field_name for the series_id. To submit immediately, set batch_size to 1 or use submit().
 
         :param str series_id: series id
         :param str field_name: name of the field
         :param str field_value: value of the field
         """
+        warnings.warn('"get_point()" is deprecated', DeprecationWarning)
+
         request = self._init_get_series_dict(series_id)
         request['fields'].update({field_name: field_value})
 
@@ -1351,7 +1439,9 @@ class BaseJob(object):
 
     def put_fields(self, series_id, fields):
         """
-        Writes the fields for the series_id.  To submit immediately, set batch_size to 1 or use submit().
+        DEPRECATED: Use write() method instead.
+
+        Writes the fields for the series_id. To submit immediately, set batch_size to 1 or use submit().
 
         :param srt series_id: series id
         :param dict fields: fields dict
@@ -1366,6 +1456,8 @@ class BaseJob(object):
 
     def remove_points(self, series_id):
         """
+        DEPRECATED: Use write(...remove_others="points") instead
+
         Sets a flag to remove all points from the series_id before the next put_point(s) call.
 
         :param series_id: series id to set the remove points flag
@@ -1374,9 +1466,11 @@ class BaseJob(object):
 
     def remove_fields(self, series_id):
         """
-        Sets a flag to remove all points from the series_id before the next put_field(s) call.
+        DEPRECATED: Use write(...remove_others="fields") instead
 
-        :param series_id: series id to set the remove points flag
+        Sets a flag to remove all fields from the series_id before the next put_field(s) call.
+
+        :param series_id: series id to set the remove fields flag
         """
         self._remove[series_id]['fields'] = True
 
@@ -1483,6 +1577,34 @@ class RemoteJob(BaseJob):
 
             bulk_data.append(tmp_dict)
 
+        for (query, reported_date), values in self._queries.items():
+            # here we will construct a BULK API call
+            if values['op'] == 'delete':
+                tmp_dict = {"type": "DELETE", "series_query": query, 'force': values['force']}
+            else:  # post
+                points = list(sorted(six.iteritems(values['points'])))
+                tmp_dict = {"type": "POST", "series_query": query}
+                if points:
+                    tmp_dict['points'] = points
+                if values['fields']:
+                    tmp_dict['fields'] = values['fields']
+
+            query_remove = self._remove_query.pop(query, None)
+
+            if query_remove:
+                if query_remove['points'] and query_remove['fields']:
+                    remove = 'all'
+                elif query_remove['points']:
+                    remove = 'points'
+                else:
+                    remove = 'fields'
+                tmp_dict['keep_only'] = remove
+
+            if reported_date is not None:
+                tmp_dict['reported_date'] = to_milli(reported_date)
+
+            bulk_data.append(tmp_dict)
+
         # we send the bulk data at once
         responses = []
         if bulk_data:
@@ -1492,7 +1614,8 @@ class RemoteJob(BaseJob):
                                                                async_mode=self._async_mode)
             finally:
                 # always flush cache
-                self._requests = {}
+                self._requests = OrderedDict()
+                self._queries = OrderedDict()
                 self._cur_batch_size = 0
         self._run_post_submit_hooks({'responses': responses, 'success': True})
         return True
@@ -1536,7 +1659,9 @@ class DryRunJob(BaseJob):
                     print_msg += ' (only {}/{} shown)'.format(self.POINTS_PRINT_LIMIT, len(points))
                 print(print_msg)
         self._remove = defaultdict(lambda: {'fields': False, 'points': False})
-        self._requests = {}
+        self._remove_query = defaultdict(lambda: {'fields': False, 'points': False})
+        self._requests = OrderedDict()
+        self._queries = OrderedDict()
 
     def delete(self, series_id, force=False):
         print('deleted series {} with force={}'.format(series_id, force))
@@ -1600,8 +1725,8 @@ class GetBulk(object):
         data = {
             '_get_type': 'get_point',
             'date_format': 'milli',
-            'df': _parse_dt(date, 'MIN'),
-            'dt': _parse_dt(date, 'MAX'),
+            'df': _parse_dt(date, 'df', 'MIN'),
+            'dt': _parse_dt(date, 'dt', 'MAX'),
             'series_id': series_id,
             'max_points': 1,
             'snapshot_job_id': snapshot_job_id,
@@ -1658,8 +1783,8 @@ class GetBulk(object):
         data = {
             '_get_type': 'get_points',
             'date_format': 'milli',
-            'df': _parse_dt(date_start, 'MIN'),
-            'dt': _parse_dt(date_finish, 'MAX'),
+            'df': _parse_dt(date_start, 'date_start', 'MIN'),
+            'dt': _parse_dt(date_finish, 'date_finish', 'MAX'),
             'max_points': max_points,
             'series_id': series_id,
             'snapshot_job_id': snapshot_job_id,
@@ -1742,8 +1867,8 @@ class GetBulk(object):
         data = {
             '_get_type': 'get_points',
             'date_format': 'milli',
-            'df': _parse_dt(df, 'MIN') if df is not None else None,
-            'dt': _parse_dt(dt, 'MAX') if dt is not None else None,
+            'df': _parse_dt(df, 'df', 'MIN') if df is not None else None,
+            'dt': _parse_dt(dt, 'dt', 'MAX') if dt is not None else None,
             'max_points': max_points,
             'query': series_query,
             'snapshot_date': to_milli(snapshot_date) if snapshot_date else None,
